class AccessDeniedError(Exception):
    def __init__(self, message: str):
        super().__init__(message)

class UserDoesNotHavePushNotificationsError(Exception):
    def __init__(self, message: str):
        super().__init__(message)

class AuthorizationRequestExpiredError(Exception):
    def __init__(self, message: str):
        super().__init__(message)



import inspect
import os
import time
from enum import Enum
from typing import Any, Awaitable, Callable, Optional, TypedDict, Union
from auth0 import Auth0Error
from auth0_ai.errors import AccessDeniedError, AuthorizationRequestExpiredError, UserDoesNotHavePushNotificationsError
from auth0_ai.token_response import TokenResponse
from auth0_ai.credentials import Credentials
from auth0_ai.authorizers.types import AuthorizerParams
from auth0.authentication.back_channel_login import BackChannelLogin
from auth0.authentication.get_token import GetToken

class CibaAuthorizerCheckResponse(Enum):
    PENDING = "pending"
    APPROVED = "approved"
    REJECTED = "rejected"
    EXPIRED = "expired"

class AuthorizeResponse(TypedDict):
    auth_req_id: str
    expires_in: int
    interval: int

class CibaCheckReponse(TypedDict):
    token: Optional[TokenResponse]
    status: CibaAuthorizerCheckResponse

class CibaAuthorizerOptions(TypedDict):
    """
    Authorize Options to start CIBA flow.

    Attributes:
        scope (str): Space-separated list of OIDC and custom API scopes.
        binding_message (Union[str, Callable[..., Awaitable[str]]]): A human-readable string to display to the user, or a function that resolves it.
        user_id (Union[str, Callable[..., Awaitable[str]]]): The user id string, or a function that resolves it.
        audience (Optional[str]): Unique identifier of the audience for an issued token.
        request_expiry (Optional[int]): To configure a custom expiry time in seconds for CIBA request, pass a number between 1 and 300.
    """
    scope: str
    binding_message: Union[str, Callable[..., Awaitable[str]]]
    user_id: Union[str, Callable[..., Awaitable[str]]]
    audience: Optional[str]
    request_expiry: Optional[int]

class CIBAAuthorizer:
    def __init__(self, options: AuthorizerParams = None):
        params = {
            "domain": (options or {}).get("domain", os.getenv("AUTH0_DOMAIN")),
            "client_id": (options or {}).get("client_id", os.getenv("AUTH0_CLIENT_ID")),
            "client_secret": (options or {}).get("client_secret", os.getenv("AUTH0_CLIENT_SECRET")),
            "client_assertion_signing_key": (options or {}).get("client_assertion_signing_key"),
            "client_assertion_signing_alg": (options or {}).get("client_assertion_signing_alg"),
            "telemetry": (options or {}).get("telemetry"),
            "timeout": (options or {}).get("timeout"),
            "protocol": (options or {}).get("protocol")
        }

        # Remove keys with None values
        params = {k: v for k, v in params.items() if v is not None}

        self.back_channel_login = BackChannelLogin(**params)
        self.get_token = GetToken(**params)

    def _ensure_openid_scope(self, scope: str) -> str:
        scopes = scope.strip().split()
        if "openid" not in scopes:
            scopes.insert(0, "openid")
        return " ".join(scopes)

    async def _start(self, params: CibaAuthorizerOptions, tool_context: Optional[Any]) -> AuthorizeResponse:
        authorize_params = {
            "scope": self._ensure_openid_scope(params.get("scope")),
            "audience": params.get("audience"),
            "request_expiry": params.get("request_expiry"),
        }

        if isinstance(params.get("user_id"), str):
            user_id = params.get("user_id")
        elif inspect.iscoroutinefunction(params.get("user_id")):
            user_id = await params.get("user_id")(tool_context)
        else:
            user_id = params.get("user_id")(tool_context)

        authorize_params["login_hint"] = f'{{ "format": "iss_sub", "iss": "https://{self.back_channel_login.domain}/", "sub": "{user_id}" }}'

        if isinstance(params.get("binding_message"), str):
            authorize_params["binding_message"] = params.get("binding_message")
        elif inspect.iscoroutinefunction(params.get("binding_message")):
            authorize_params["binding_message"] = await params.get("binding_message")(tool_context)
        else:
            authorize_params["binding_message"] = params.get("binding_message")(tool_context)
        
        response = self.back_channel_login.back_channel_login(**authorize_params)
        return AuthorizeResponse(
            auth_req_id=response["auth_req_id"],
            expires_in=response["expires_in"],
            interval=response["interval"],
        )

    async def _check(self, auth_req_id: str) -> CibaCheckReponse:
        response = CibaCheckReponse(status=CibaAuthorizerCheckResponse.PENDING)

        try:
            result = self.get_token.backchannel_login(auth_req_id=auth_req_id)
            response["status"] = CibaAuthorizerCheckResponse.APPROVED
            response["token"] = {
                "access_token": result["access_token"],
                "id_token": result["id_token"],
                "expires_in": result["expires_in"],
                "scope": result["scope"],
                "refresh_token": result.get("refresh_token"),
                "token_type": result.get("token_type"),
            }
        except Auth0Error as e:
            if e.error_code == "invalid_request":
                response["status"] = CibaAuthorizerCheckResponse.EXPIRED
            elif e.error_code == "access_denied":
                response["status"] = CibaAuthorizerCheckResponse.REJECTED
            elif e.error_code == "authorization_pending":
                response["status"] = CibaAuthorizerCheckResponse.PENDING
        
        return response
    
    async def poll(self, params: AuthorizeResponse) -> Credentials:
        start_time = time.time()

        while time.time() - start_time < params.get("expires_in"):
            try:
                response = self.get_token.backchannel_login(auth_req_id=params.get("auth_req_id"))
                return Credentials(
                    access_token={
                        "type": response.get("token_type", "bearer"),
                        "value": response["access_token"],
                    }
                )
            except Auth0Error as e:
                if e.error_code == "invalid_request":
                    raise UserDoesNotHavePushNotificationsError(e.message)
                elif e.error_code == "access_denied":
                    raise AccessDeniedError(e.message)
                elif e.error_code == "authorization_pending":
                    time.sleep(params.get("interval"))
                else:
                    raise AuthorizationRequestExpiredError("Authorization request expired")

    @staticmethod
    async def start(options: CibaAuthorizerOptions, params: AuthorizerParams = None, tool_context: Any = None) -> AuthorizeResponse:
        authorizer = CIBAAuthorizer(params)
        return await authorizer._start(options, tool_context)

    @staticmethod
    async def check(auth_req_id: str, params: AuthorizerParams = None) -> CibaCheckReponse:
        authorizer = CIBAAuthorizer(params)
        return await authorizer._check(auth_req_id)



from typing import TypedDict, Optional

class TokenResponse(TypedDict):
    access_token: str
    id_token: str
    expires_in: int
    scope: str
    refresh_token: Optional[str]
    token_type: Optional[str]



from typing import TypedDict, Optional

class Credential(TypedDict):
    type: str
    value: str

class Credentials(TypedDict):
    access_token: Credential
    id_token: Optional[Credential]
    refresh_token: Optional[Credential]



from typing import Dict, Any
from .auth0_interrupt import Auth0Interrupt

class FederatedConnectionInterrupt(Auth0Interrupt):
    """
    Error thrown when a tool call requires an access token for an external service.

    Throw this error if the service returns Unauthorized for the current access token.
    """

    code = "FEDERATED_CONNECTION_ERROR"

    def __init__(self, message: str, connection: str, scopes: list[str], required_scopes: list[str]):
        """
        Initializes a FederatedConnectionInterrupt instance.

        Args:
            message (str): Error message describing the reason for the interrupt.
            connection (str): The Auth0 connection name.
            scopes (list[str]): The scopes required to access the external service as stated in the authorizer.
            required_scopes (list[str]): The union between the current scopes of the Access Token plus the required scopes.
                                         This is the list of scopes that will be used to request a new Access Token.
        """
        super().__init__(message, self.code)
        self.connection = connection
        self.scopes = scopes
        self.required_scopes = required_scopes

    def to_json(self) -> Dict[str, Any]:
        """
        Serializes the interrupt to a JSON object.
        """
        return {
            **super().to_json(),
            "connection": self.connection,
            "scopes": self.scopes,
            "requiredScopes": self.required_scopes,
        }


class FederatedConnectionError(Exception):
    """
    Error thrown when a tool call requires an access token for an external service.

    The authorizer will automatically convert this class of error to FederatedConnectionInterrupt.
    """

    def __init__(self, message: str):
        """
        Initializes a FederatedConnectionError instance.

        Args:
            message (str): Error message describing the reason for the error.
        """
        super().__init__(message)



import asyncio
import contextvars
import inspect
import os
from contextlib import asynccontextmanager
from typing import Awaitable, Callable, Generic, Optional, Any, TypedDict, Union
from auth0.authentication.get_token import GetToken
from .types import AuthorizerParams, AuthorizerToolParameter, ToolInput
from ..token_response import TokenResponse
from ..interrupts.auth0_interrupt import Auth0Interrupt
from ..interrupts.federated_connection_interrupt import FederatedConnectionError, FederatedConnectionInterrupt

class AsyncStorageValue(TypedDict, total=False):
    context: Any
    connection: str
    scopes: list[str]
    access_token: Optional[str]
    current_scopes: Optional[list[str]]

_local_storage: contextvars.ContextVar[Optional[AsyncStorageValue]] = contextvars.ContextVar("local_storage", default=None)

def _get_local_storage() -> AsyncStorageValue:
    store = _local_storage.get()
    if store is None:
        raise RuntimeError("The tool must be wrapped with the with_federated_connection function.")
    return store

def _update_local_storage(data: AsyncStorageValue) -> None:
    store = _get_local_storage()
    updated = store.copy()
    updated.update(data)
    _local_storage.set(updated)

@asynccontextmanager
async def _run_with_local_storage(data: AsyncStorageValue):
    if _local_storage.get() is not None:
        raise RuntimeError("Cannot nest tool calls that require federated connection authorization.")
    token = _local_storage.set(data)
    try:
        yield
    finally:
        _local_storage.reset(token)

def get_access_token_for_connection() -> str | None:
    store = _get_local_storage()
    return store.get("access_token")

class FederatedConnectionAuthorizerParams(Generic[ToolInput]):
    def __init__(
        self,
        scopes: Union[
            AuthorizerToolParameter[ToolInput, list[str]],
            Callable[ToolInput, Union[list[str], Awaitable[list[str]]]],
            list[str],
        ],
        connection: Union[
            AuthorizerToolParameter[ToolInput, str],
            Callable[ToolInput, Union[str, Awaitable[str]]],
            str,
        ],
        refresh_token: Optional[Union[
            AuthorizerToolParameter[ToolInput, str | None],
            Callable[ToolInput, Union[str | None, Awaitable[str | None]]],
            str | None,
        ]] = None,
        access_token: Optional[Union[
            AuthorizerToolParameter[ToolInput, TokenResponse | None],
            Callable[ToolInput, Union[TokenResponse | None, Awaitable[TokenResponse | None]]],
            TokenResponse | None
        ]] = None,
    ):
        def wrap(val, result_type):
            if isinstance(val, AuthorizerToolParameter):
                return val
            return AuthorizerToolParameter[ToolInput, result_type](val)
       
        self.scopes = wrap(scopes, list[str])
        self.connection = wrap(connection, str)
        self.refresh_token = wrap(refresh_token, str | None)
        self.access_token = wrap(access_token, TokenResponse | None)

class FederatedConnectionAuthorizerBase(Generic[ToolInput]):
    def __init__(
        self,
        options: FederatedConnectionAuthorizerParams[ToolInput],
        config: AuthorizerParams = None,
    ):
        self.options = options
        auth0 = {
            "domain": (config or {}).get("domain", os.getenv("AUTH0_DOMAIN")),
            "client_id": (config or {}).get("client_id", os.getenv("AUTH0_CLIENT_ID")),
            "client_secret": (config or {}).get("client_secret", os.getenv("AUTH0_CLIENT_SECRET")),
            "client_assertion_signing_key": (config or {}).get("client_assertion_signing_key"),
            "client_assertion_signing_alg": (config or {}).get("client_assertion_signing_alg"),
            "telemetry": (config or {}).get("telemetry"),
            "timeout": (config or {}).get("timeout"),
            "protocol": (config or {}).get("protocol")
        }

        # Remove keys with None values
        auth0 = {k: v for k, v in auth0.items() if v is not None}
        self.get_token = GetToken(**auth0)

        # Ensure either refreshToken or accessToken is provided
        if options.refresh_token.value is None and options.access_token.value is None:
            raise ValueError("Either refresh_token or access_token must be provided to initialize the Authorizer.")
        
        if options.refresh_token.value is not None and options.access_token.value is not None:
            raise ValueError("Only one of refresh_token or access_token can be provided to initialize the Authorizer.")
    
    def _handle_authorization_interrupts(self, err: Auth0Interrupt) -> None:
        raise err
    
    def validate_token(self, token_response: Optional[TokenResponse] = None):
        store = _get_local_storage()
        scopes = store["scopes"]
        connection = store["connection"]
        
        if token_response is None:
            raise FederatedConnectionInterrupt(
                f"Authorization required to access the Federated Connection API: {connection}",
                connection,
                scopes,
                scopes
            )

        current_scopes = token_response["scope"].split(" ") if token_response["scope"] else []
        missing_scopes = [s for s in scopes if s not in current_scopes]
        _update_local_storage({"current_scopes": current_scopes})
        
        if missing_scopes:
            raise FederatedConnectionInterrupt(
                f"Authorization required to access the Federated Connection API: {connection}. Missing scopes: {', '.join(missing_scopes)}",
                connection,
                scopes,
                current_scopes + scopes
            )

    async def get_access_token_impl(self, *args: ToolInput.args, **kwargs: ToolInput.kwargs) -> TokenResponse | None:
        store = _get_local_storage()
        
        connection = store["connection"]
        subject_token = await self.get_refresh_token(*args, **kwargs)
        if not subject_token:
            return None
        
        # TODO: replace once the auth0-python sdk supports this feature
        response = self.get_token.authenticated_post(
            f"{self.get_token.protocol}://{self.get_token.domain}/oauth/token",
            data={
                "grant_type": "urn:auth0:params:oauth:grant-type:token-exchange:federated-connection-access-token",
                "client_id": self.get_token.client_id,
                "subject_token_type": "urn:ietf:params:oauth:token-type:refresh_token",
                "subject_token": subject_token,
                "connection": connection,
                "requested_token_type": "http://auth0.com/oauth/token-type/federated-connection-access-token",
            },
        )
        
        return TokenResponse(**response)
    
    async def get_access_token(self, *args: ToolInput.args, **kwargs: ToolInput.kwargs) -> TokenResponse | None:
        if callable(self.options.refresh_token.value) or asyncio.iscoroutinefunction(self.options.refresh_token.value):
            token_response = await self.get_access_token_impl(*args, **kwargs)
        else:
            token_response = await self.options.access_token.resolve(*args, **kwargs)
        
        self.validate_token(token_response)
        return token_response
    
    async def get_refresh_token(self, *args: ToolInput.args, **kwargs: ToolInput.kwargs):
        return await self.options.refresh_token.resolve(*args, **kwargs)
    
    def protect(
        self,
        get_context: Callable[ToolInput, any],
        execute: Callable[ToolInput, any]
    ) -> Callable[ToolInput, any]:
        async def wrapped_execute(*args: ToolInput.args, **kwargs: ToolInput.kwargs):
            store = {
                "context": get_context(*args, **kwargs),
                "scopes": await self.options.scopes.resolve(*args, **kwargs),
                "connection": await self.options.connection.resolve(*args, **kwargs)
            }

            async with _run_with_local_storage(store):
                try:
                    token_response = await self.get_access_token(*args, **kwargs)
                    _update_local_storage({"access_token": token_response["access_token"]})

                    if inspect.iscoroutinefunction(execute):
                        return await execute(*args, **kwargs)
                    else:
                        return execute(*args, **kwargs)
                except FederatedConnectionError as err:
                    interrupt = FederatedConnectionInterrupt(
                        str(err),
                        store["connection"],
                        store["scopes"],
                        store["scopes"]
                    )
                    return self._handle_authorization_interrupts(interrupt)
                except Auth0Interrupt as err:
                    return self._handle_authorization_interrupts(err)
                except Exception as err:
                    raise err
        
        return wrapped_execute



from typing import Type, TypeVar, Dict, Any

T = TypeVar("T", bound="Auth0Interrupt")

class Auth0Interrupt(Exception):
    def __init__(self, message: str, code: str):
        super().__init__(message)
        self.code = code

    def to_json(self) -> Dict[str, Any]:
        """
        Serialize the interrupt to a JSON object.
        """
        return {
            key: value for key, value in self.__dict__.items()
        } | {"message": self.args[0]}

    @classmethod
    def is_interrupt(cls: Type[T], interrupt: Any) -> bool:
        """
        Checks if an interrupt is of a specific type asserting its data component.
        """
        return (
            isinstance(interrupt, dict) and
            interrupt.__class__.__name__ == "Auth0Interrupt" and
            (not hasattr(cls, "code") or interrupt.get("code") == getattr(cls, "code", None))
        )
